\
\ Code for the DCC generator.
\
\ $ProjectVersion: 0.196 $
\

include picisr.fs
include piceeprom.fs
include multitasker.fs
include libfetch.fs
include libstore.fs
include libnibble.fs

\ ----------------------------------------------------------------------
\ Pinout used in this program
\ ----------------------------------------------------------------------

0 pin-b dcc-output
1 pin-b led-empty
2 pin-b led-error

\ ----------------------------------------------------------------------
\ Store and fetch from memory banks
\ ----------------------------------------------------------------------

code bank-get-prepare ( addr -- x )
    fsr ,w movf               \ Get stack pointer
    tmp1 movwf                \ and save it
    loadw                     \ Get address,
    fsr movwf                 \ prepare to read value
    return
end-code

code bank-get-terminate ( x -- n )
    tmp2 movwf                \ Save read value
    tmp1 ,w movf              \ Get old stack pointer
    fsr movwf                 \ and restore it
    tmp2 ,w movf              \ Get saved value
    indf movwf                \ Write the read value
    return
end-code

code bank1-get ( addr -- n )
    bank-get-prepare call
    reg-rp0 bsf              \ Choose bank 1
    indf ,w movf             \ read value
    reg-rp0 bcf              \ Restore bank 0
    bank-get-terminate goto
end-code

code bank2-get ( addr -- n )
    bank-get-prepare call
    reg-rp1 bsf               \ Choose bank 2
    indf ,w movf              \ read value
    reg-rp1 bcf               \ Restore bank 0
    bank-get-terminate goto
end-code

code bank-set-prepare ( v addr -- )
    popw                      \ Get address
    tmp1 movwf                \ and save it
    popw                      \ Get value
    tmp2 movwf                \ and save it
    fsr ,w movf               \ Get stack pointer
    tmp3 movwf                \ and save it
    tmp1 ,w movf              \ Get address
    fsr movwf                 \ and prepare to write value
    tmp2 ,w movf              \ Get value
    return
end-code

code bank1-set ( v addr -- )
    bank-set-prepare call
    reg-rp0 bsf               \ Choose bank 1
    indf movwf                \ Write value
    reg-rp0 bcf               \ Restore bank 0
    tmp3 ,w movf              \ Get old stack pointer
    fsr movwf                 \ and restore it
    return
end-code

code bank2-set ( v addr -- )
    bank-set-prepare call
    reg-rp1 bsf               \ Choose bank 2
    indf movwf                \ Write value
    reg-rp1 bcf               \ Restore bank 0
    tmp3 ,w movf              \ Get old stack pointer
    fsr movwf                 \ and restore it
    return
end-code

variable current-bank1        \ True if working in bank 1, false if in bank 2

: bank1 ( -- ) 1 current-bank1 ! ;
: bank2 ( -- ) 0 current-bank1 ! ;

: bank-get ( addr -- n ) current-bank1 @ if bank1-get exit then bank2-get ;
: bank-set ( v addr -- ) current-bank1 @ if bank1-set exit then bank2-set ;

variable src
variable dst

\ Increment src and dst

: inc-vars ( -- ) 1 src +! 1 dst +! ;

\ Copy one byte from src to dst in the selected bank

: to-bank ( -- ) src @ @ dst @ bank-set ;

\ Copy one byte from src in the selected bank to dst

: from-bank ( -- ) src @ bank-get dst @ ! ;

\ ----------------------------------------------------------------------
\ Serial port handling
\ ----------------------------------------------------------------------

\ When a task wants to send something, it has to:
\   - wait for tr-length to be zero
\   - write address (in EEPROM) in tr-eeaddr
\   - write length in tr-length
\
\ Note that no lock is necessary as cooperative multitasking ensures that
\ the whole operation is atomic.

variable tr-eeaddr
variable tr-length

task : serial-transmit ( -- )
    begin
	\ Wait until something is ready to be transmitted (yield at least
	\ once per character)
	begin yield tr-length @ until
	\ Wait for the authorization to transmit
	begin txif bit-set? /while yield repeat
	\ Transmit data
	tr-eeaddr @ ee@ txreg !
	\ One character less to transmit
	1 tr-eeaddr +! 1 tr-length -!
    again
;

macro

: key ( -- c ) begin rcif bit-set? /while yield repeat rcreg @ ;
    
target

variable cmd-length
eecreate cmd 7 allot
variable nibble

task : serial-receive ( -- )
    begin
	\ Wait for '!' which indicates a command start
	key [char] ! - if recurse exit then
	\ Get the command (up to 7 hex bytes, followed by ".")
	0 nibble ! cmd
	begin
	    key dup [char] ! - /if 2drop recurse exit then
	    dup [char] . - while
	    hex>nibble
	    nibble @ if
		\ Wait for eeprom with yield if needed
		begin wr bit-set? while yield repeat
		\ Shift high nibble and add low nibble
		nibble @ 4 lshift or
		\ Store in EEPROM
		eedata ! dup eeadr ! eeprom-write
		\ Reset nibble and increment address
		0 nibble ! 1+
	    else
		\ In case nibble is zero, the highest bit is set
		80 or nibble !
	    then
	repeat drop cmd - cmd-length !
	\ XXXXX nothing done with the command yet
	\ Send acknowledgment
	begin tr-length @ while yield repeat
	l" Ok" tr-length ! tr-eeaddr !
    again
;

\ ----------------------------------------------------------------------
\ Interrupt driven DCC pulse generator
\ ----------------------------------------------------------------------

\ set-timer will wait for 3.2*(256-n)-0.2 s for a 20MhZ oscillator and a
\ 16 prescaler. For ones, use n=238 (57.4s => 1.03% error). For zero, use
\ values between n=0 (819.0s) and n=224 (102.2s). If necessary, the range
\ will be adjusted by testing the timer a different

macro
: set-timer ( n -- ) tmr0 +! ;
target

\ delay contains the duration of the next low pulse when the high pulse
\ is in effect, and 0 otherwise
\ Warning: those words (pulse, bit-zero and bit-one) terminate with a call
\ to isr-exit, thus returning from interrupt

variable delay

\ pulse and half-pulse work the reverse way around: they set the timer
\ to the proper value, then call the idle-loop to do other tasks, then
\ wait for the timer to expire and immediately change the DCC output
\ state.  This way, it is permitted to take some time (up to 40s,
\ i.e. around 200 instructions) to handle secondary tasks. More
\ complex tasks can be handled if the zero flag is true (in which case
\ we have up to 80s, i.e. around 400 instructions).

: half-pulse ( -- ) delay @ set-timer ;          \ Set timer to pulse value

\ isr-exit is a word doing isr-restore-return

code isr-exit ( -- )
    isr-restore-return
end-code

:: pulse ( duration -- ) delay ! half-pulse isr-exit ;

: bit-zero ( -- ) $64 pulse ;
: bit-one ( -- )  $ee pulse ;

\ Data buffer to be transmitted. As soon as the length is zero, it is
\ possible to write another two data bytes, then the length. Since the
\ DCC pulse generator moves the data in its internal buffer, a task
\ has virtually the length of a byte plus a stop or a start bit to
\ prepare its data to be sent.

variable dcc-high
variable dcc-low
variable dcc-length

\ Data being transmitted. We implement a 16 bits shift register, left-aligned,
\ so that preamble, start and stop bits can be easily implemented.

variable data-high
variable data-low
variable data-length

\ isr-tmr0 is called when there is a timer overflow interrupt
\ This word is very long to avoid return stack depth exhaustion, but very
\ thoroughly commented. It terminates with a call to isr-exit to properly
\ exit from interrupt

isr : isr-tmr0 ( -- )
    t0if bit-clr                        \ Acknowledge interrupt
    delay @ if
	dcc-output low                  \ Drive DCC output low
	half-pulse                      \ Reset timer and clear watchdog
	0 delay !                       \ Note low pulse in progress
	isr-exit exit                   \ Properly terminate isr
    then
    dcc-output high                     \ Drive DCC output high
    1 data-length -!                    \ Decrement data length
    c-if led-error high then            \ Signal error if no data ready
    ]asm
    data-low ,f rlf                     \ Shift data bytes one bit
    data-high ,f rlf                    \ left through carry
    asm[
    c-if bit-one exit then              \ Transmit one
    bit-zero                            \ Transmit zero
;

\ This task feeds the interrupt routine with data (double buffering)

task : feeder ( -- )
    begin
	\ Wait for the transmit buffer to be empty
	begin data-length @ while yield repeat
	dcc-length @ if
	    \ If there is data to transmit, copy it
	    dcc-high @ data-high ! dcc-low @ data-low !
	    dcc-length @ data-length ! 0 dcc-length !
	    led-empty low
	else
	    \ No data, transmit a zero bit
	    0 data-high ! 1 data-length !
	    led-empty high
	then
    again
;

\ ----------------------------------------------------------------------
\ DCC data manipulation
\ ----------------------------------------------------------------------

variable checksum

\ Wait for the transmission slot to be available

: wait-for-transmission ( -- )
    begin dcc-length @ while repeat ;

\ Transmit a preamble (10 bits to 1) and a start bit (one bit to 0)

: preamble ( -- )
    wait-for-transmission
    0 checksum !
    $ff dcc-high ! $c0 dcc-low ! $b dcc-length ! ;

\ send-byte sends a byte followed by a start bit. send-final-byte ends
\ the packet with a stop bit after the byte. send-checksum sends the checksum
\ with a stop bit.

: finish-byte ( b -- ) dup checksum xor! dcc-high ! 9 dcc-length ! ;

: send-byte ( b -- )
    wait-for-transmission $00 dcc-low ! finish-byte ;

: send-final-byte ( b -- )
    wait-for-transmission $80 dcc-low ! finish-byte ;

: send-checksum ( -- ) checksum @ send-final-byte ;

\ Short loco speed command

: send-loco-speed ( s n -- )
    preamble send-byte send-byte send-checksum ;

\ ----------------------------------------------------------------------
\ Initialisation
\ ----------------------------------------------------------------------

: init-portb ( -- )
    $00 portb !        \ Clear output latches
    $f8 trisb !        \ Ports 0, 1 and 2 as output
    \ Interrupts should be configured here
;

: init-timer ( -- )
    $03 option_reg !             \ CLKOUT, prescaler to timer 0, prescaler=16
    t0ie bit-set                 \ Enable timer 0 overflow interrupt
    $f0 tmr0 !                   \ Do not wait too long before first interrupt
    ;

: init-bank1 ( -- ) 0 a0 bank1-set ;

: init-serial ( -- )
    0 tr-length !
    $90 rcsta !                  \ Serial-enable 8bits-rx continuous-receive
    $24 txsta !                  \ Asynchronous transmit high-speed generator
    $81 spbrg !                  \ 9600 bauds with BGRH high at 20MhZ
    portc 6 >output              \ Port C6 is TX
    l" Welcome to the DCC generator" tr-length ! tr-eeaddr !
;

: init-dcc ( -- )
    0 dcc-length !               \ No data being transmitted
    0 data-high !                \ 1 zero queued for transmission
    1 data-length !
;

: init ( -- )
    init-dcc init-portb init-bank1 init-serial init-timer enable-interrupts
;

\ ----------------------------------------------------------------------
\ Main program and main loop
\ ----------------------------------------------------------------------

main
: main ( -- ) init multitasker ;

\ ----------------------------------------------------------------------
\ Configuration word
\ ----------------------------------------------------------------------

fosc-hs set-fosc    \ High-speed oscillator
